#ifndef __HEADER_H__
#define __HEADER_H__

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>

/**
 * DNS Packet Header class
 *
 * This class handles parsing and constructing DNS Packet Headers as defined
 * by section 4.1.1 of RFC1035.
 * 
 *  DNS header format - RFC1035 section 4.1.1
 *  DNS header format - RFC4035 section 3.2
 *
 *      0  1  2  3  4  5  6  7  8  9  0  1  2  3  4  5
 *    +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
 *    |                      ID                       |
 *    +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
 *    |QR|   Opcode  |AA|TC|RD|RA| Z|AD|CD|   RCODE   |
 *    +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
 *    |                    QDCOUNT                    |
 *    +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
 *    |                    ANCOUNT                    |
 *    +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
 *    |                    NSCOUNT                    |
 *    +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
 *    |                    ARCOUNT                    |
 *    +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
 *
 * @category Networking
 * @package  Net_DNS2
 * @author   Mike Pultz <mike@mikepultz.com>
 * @license  http://www.opensource.org/licenses/bsd-license.php  BSD License
 * @link     http://pear.php.net/package/Net_DNS2
 *
 */

typedef struct {
	uint16_t id;         // 16 bit - identifier
	uint8_t qr;         //  1 bit - 0 = query, 1 = response
	uint8_t opcode;     //  4 bit - op code
	uint8_t aa;         //  1 bit - Authoritative Answer
	uint8_t tc;         //  1 bit - TrunCation
	uint8_t rd;         //  1 bit - Recursion Desired
	uint8_t ra;         //  1 bit - Recursion Available
	uint8_t z;          //  1 bit - Reserved
	uint8_t ad;         //  1 bit - Authentic Data (RFC4035)
	uint8_t cd;         //  1 bit - Checking Disabled (RFC4035)
	uint8_t rcode;      //  4 bit - Response code
	uint16_t qdcount;    // 16 bit - entries in the question section
	uint16_t ancount;    // 16 bit - resource records in the answer section
	uint16_t nscount;    // 16 bit - name server rr in the authority records section
	uint16_t arcount;    // 16 bit - rr's in the additional records section
} DNSHeader;

void dns_header_init (DNSHeader *header);

#endif /* __HEADER_H__ */
